@@ -1,14 +1,14 @@
 #  You may distribute under the terms of either the GNU General Public License
 #  or the Artistic License (the same terms as Perl itself)
 #
-#  (C) Paul Evans, 2016-2018 -- leonerd@leonerd.org.uk
+#  (C) Paul Evans, 2016-2020 -- leonerd@leonerd.org.uk
 
 package Net::Prometheus;
 
 use strict;
 use warnings;
 
-our $VERSION = '0.07';
+our $VERSION = '0.11';
 
 use Carp;
 
@@ -19,12 +19,13 @@ use Net::Prometheus::Counter;
 use Net::Prometheus::Summary;
 use Net::Prometheus::Histogram;
 
+use Net::Prometheus::Registry;
+
 use Net::Prometheus::ProcessCollector;
+use Net::Prometheus::PerlCollector;
 
 use Net::Prometheus::Types qw( MetricSamples );
 
-use constant HAVE_PERLCOLLECTOR => defined eval { require Net::Prometheus::PerlCollector };
-
 =head1 NAME
 
 C<Net::Prometheus> - export monitoring metrics for F<prometheus>
@@ -64,6 +65,17 @@ program can embed in its own structure to provide the results, or the
 application can generate a plain-text result directly and serve them by its
 own means.
 
+=head2 Metrics::Any
+
+For more flexibility of metrics reporting, other modules may wish to use
+L<Metrics::Any> as an abstraction interface instead of directly using this
+API.
+
+By using C<Metrics::Any> instead, the module does not directly depend on
+C<Net::Prometheus>, and in addition program ultimately using the module gets
+the flexibility to use Prometheus (via L<Metrics::Any::Adapter::Prometheus>)
+or use another reporting system via a different adapter.
+
 =cut
 
 =head1 CONSTRUCTOR
@@ -90,7 +102,7 @@ collector will be loaded by default.
 
 If present and true, this instance will not load perl-specific collector from
 L<Net::Prometheus::PerlCollector>. If absent or false this collector is loaded
-if the module is installed and useable.
+by default.
 
 These two options are provided for testing purposes, or for specific use-cases
 where such features are not required. Usually it's best just to leave these
@@ -106,7 +118,7 @@ sub new
    my %args = @_;
 
    my $self = bless {
-      collectors => [],
+      registry => Net::Prometheus::Registry->new,
    }, $class;
 
    if( not $args{disable_process_collector} and
@@ -114,7 +126,7 @@ sub new
       $self->register( $process_collector );
    }
 
-   if( not $args{disable_perl_collector} and HAVE_PERLCOLLECTOR ) {
+   if( not $args{disable_perl_collector} ) {
       $self->register( Net::Prometheus::PerlCollector->new );
    }
 
@@ -139,11 +151,7 @@ sub register
    my $self = shift;
    my ( $collector ) = @_;
 
-   # TODO: ban duplicate registration
-
-   push @{ $self->{collectors} }, $collector;
-
-   return $collector;
+   return $self->{registry}->register( $collector );
 }
 
 =head2 unregister
@@ -159,13 +167,7 @@ sub unregister
    my $self = shift;
    my ( $collector ) = @_;
 
-   my $found;
-   @{ $self->{collectors} } = grep {
-      not( $_ == $collector and ++$found )
-   } @{ $self->{collectors} };
-
-   $found or
-      croak "No such collector";
+   return $self->{registry}->unregister( $collector );
 }
 
 =head2 new_gauge
@@ -277,7 +279,7 @@ sub new_metricgroup
 
 =head2 collect
 
-   @metricsamples = $prometheus->collect
+   @metricsamples = $prometheus->collect( $opts )
 
 Returns a list of L<Net::Prometheus::Types/MetricSamples> obtained from all
 of the currently-registered collectors.
@@ -287,12 +289,13 @@ of the currently-registered collectors.
 sub collect
 {
    my $self = shift;
+   my ( $opts ) = @_;
 
-   my $collectors = $self->{collectors};
+   $opts //= {};
 
    my %samples_by_name;
-   foreach my $collector ( @{ $collectors } ) {
-      push @{ $samples_by_name{ $_->fullname } }, $_ for $collector->collect;
+   foreach my $collector ( $self->{registry}->collectors, Net::Prometheus::Registry->collectors ) {
+      push @{ $samples_by_name{ $_->fullname } }, $_ for $collector->collect( $opts );
    }
 
    return map {
@@ -314,6 +317,11 @@ sub collect
 Returns a string in the Prometheus text exposition format containing the
 current values of all the registered metrics.
 
+   $str = $prometheus->render( { options => "for collectors" } )
+
+An optional HASH reference may be provided; if so it will be passed into the
+C<collect> method of every registered collector.
+
 =cut
 
 sub _render_label_value
@@ -340,6 +348,7 @@ sub _render_labels
 sub render
 {
    my $self = shift;
+   my ( $opts ) = @_;
 
    return join "", map {
       my $metricsamples = $_;
@@ -359,7 +368,75 @@ sub render
             _render_labels( $sample->labels ),
             $sample->value
       } @{ $metricsamples->samples }
-   } $self->collect;
+   } $self->collect( $opts );
+}
+
+=head2 handle
+
+   $response = $prometheus->handle( $request )
+
+Given an HTTP request in an L<HTTP::Request> instance, renders the metrics in
+response to it and returns an L<HTTP::Response> instance.
+
+This application will respond to any C<GET> request, and reject requests for
+any other method. If a query string is present on the URI it will be parsed
+for collector options to pass into the L</render> method.
+
+This method is useful for integrating metrics into an existing HTTP server
+application which uses these objects. For example:
+
+   my $prometheus = Net::Prometheus->new;
+
+   sub serve_request
+   {
+      my ( $request ) = @_;
+
+      if( $request->uri->path eq "/metrics" ) {
+         return $prometheus->handle( $request );
+      }
+
+      ...
+   }
+
+=cut
+
+# Some handy pseudomethods to make working on HTTP::Response less painful
+my $set_header = sub {
+   my $resp = shift;
+   $resp->header( @_ );
+   $resp;
+};
+my $set_content = sub {
+   my $resp = shift;
+   $resp->content( @_ );
+   $resp;
+};
+my $fix_content_length = sub {
+   my $resp = shift;
+   $resp->content_length or $resp->content_length( length $resp->content );
+   $resp;
+};
+
+sub handle
+{
+   my $self = shift;
+   my ( $request ) = @_;
+
+   require HTTP::Response;
+
+   $request->method eq "GET" or return
+      HTTP::Response->new( 405 )
+         ->$set_header( Content_Type => "text/plain" )
+         ->$set_content( "Method " . $request->method . " not supported" )
+         ->$fix_content_length;
+
+   my $opts;
+   $opts = { $request->uri->query_form } if length $request->uri->query;
+
+   return HTTP::Response->new( 200 )
+      ->$set_header( Content_Type => "text/plain; version=0.0.4; charset=utf-8" )
+      ->$set_content( $self->render( $opts ) )
+      ->$fix_content_length;
 }
 
 =head2 psgi_app
@@ -371,7 +448,20 @@ will render the metrics in the Prometheus text exposition format, suitable for
 scraping by the Prometheus collector.
 
 This application will respond to any C<GET> request, and reject requests for
-any other method.
+any other method. If a C<QUERY_STRING> is present in the environment it will
+be parsed for collector options to pass into the L</render> method.
+
+This method is useful for integrating metrics into an existing HTTP server
+application which is uses or is based on PSGI. For example:
+
+   use Plack::Builder;
+
+   my $prometheus = Net::Prometheus::->new;
+
+   builder {
+      mount "/metrics" => $prometheus->psgi_app;
+      ...
+   }
 
 =cut
 
@@ -379,6 +469,8 @@ sub psgi_app
 {
    my $self = shift;
 
+   require URI;
+
    return sub {
       my $env = shift;
       my $method = $env->{REQUEST_METHOD};
@@ -389,14 +481,76 @@ sub psgi_app
          [ "Method $method not supported" ],
       ];
 
+      my $opts;
+      if( defined $env->{QUERY_STRING} ) {
+         $opts = +{ URI->new( "?$env->{QUERY_STRING}", "http" )->query_form };
+      }
+
       return [
          200,
          [ "Content-Type" => "text/plain; version=0.0.4; charset=utf-8" ],
-         [ $self->render ],
+         [ $self->render( $opts ) ],
       ];
    };
 }
 
+=head2 export_to_IO_Async
+
+   $prometheus->export_to_IO_Async( $loop, %args )
+
+Performs the necessary steps to create an HTTP server for exporting metrics
+over HTTP via L<IO::Async>. This will involve creating a new
+L<Net::Async::HTTP::Server> instance added to the loop.
+
+This new server will listen on its own port number for any incoming request,
+and will serve metrics regardless of path.
+
+Note this should only be used in applications that don't otherwise have an
+HTTP server, such as self-contained monitoring exporters or exporting metrics
+as a side-effect of other activity. For existing HTTP server applications it
+is better to integrate with the existing request/response processing of the
+application, such as by using the L</handle> or L</psgi_app> methods.
+
+Takes the following named arguments:
+
+=over 4
+
+=item port => INT
+
+Port number on which to listen for incoming HTTP requests.
+
+=back
+
+=cut
+
+sub export_to_IO_Async
+{
+   my $self = shift;
+   my ( $loop, %args ) = @_;
+
+   require IO::Async::Loop;
+   require Net::Async::HTTP::Server;
+
+   $loop //= IO::Async::Loop->new;
+
+   my $httpserver = Net::Async::HTTP::Server->new(
+      on_request => sub {
+         my $httpserver = shift;
+         my ( $req ) = @_;
+
+         $req->respond( $self->handle( $req->as_http_request ) );
+      },
+   );
+
+   $loop->add( $httpserver );
+
+   # Yes this is a blocking call
+   $httpserver->listen(
+      socktype => "stream",
+      service  => $args{port},
+   )->get;
+}
+
 {
    package
       Net::Prometheus::_MetricGroup;
@@ -433,13 +587,19 @@ L</render> method. A collector can be any object instance that has a method
 called C<collect>, which when invoked is passed no arguments and expected to
 return a list of L<Net::Prometheus::Types/MetricSamples> structures.
 
-   @metricsamples = $collector->collect
+   @metricsamples = $collector->collect( $opts )
 
 The L<Net::Prometheus::Metric> class is already a valid collector (and hence,
 so too are the individual metric type subclasses). This interface allows the
 creation of new custom collector objects, that more directly collect
 information to be exported.
 
+Collectors might choose to behave differently in the presence of some
+specifically-named option; typically to provide extra detail not normally
+provided (maybe at the expense of extra processing time to calculate it).
+Collectors must not complain about the presence of unrecognised options; the
+hash is shared among all potential collectors.
+
 =cut
 
 =head1 TODO
@@ -448,11 +608,12 @@ information to be exported.
 
 =item *
 
-Split Registry out from toplevel instance.
+Histogram/Summary 'start_timer' support
 
 =item *
 
-Write some actual example programs.
+Add other C<export_to_*> methods for other event systems and HTTP-serving
+frameworks, e.g. L<Mojo>.
 
 =back
 
